from wendelin.bigarray.array_zodb import ZBigArray
import transaction
import numpy as np
import pandas as pd
from numpy import float64, dtype, cumsum, sin, uint8
from DateTime import DateTime
from time import time
from contextlib import contextmanager
import sys
import psutil
import os
import cProfile
import pstats


def bigger_than_memory(self):
  message_list = []
  message_list.extend(bigger_than_memory_write(self))
  message_list.extend(bigger_than_memory_read(self))
  return message_list
  
def bigger_than_memory_write(self, out_of_core_index=False):
  message_list = []
  message_list.append('Write start: ' + get_process_memory_usage())

  #ram_nbytes = psutil.virtual_memory().total
  #big_array_len = (1*ram_nbytes) // big_array_dtype.itemsize
  _, array_schema = get_field_names_and_schema()
                
  big_array = create_zbigarray((360000,), dtype=array_schema)
  big_index = create_zbigarray((1, 360000), dtype(float64)) if out_of_core_index else np.arange(360000)
  
  root = self.getPhysicalRoot()
  
  if out_of_core_index:
    big_array, big_index = store_arrays(root, [big_array, 'big_array'], [big_index, 'big_index'])
  else:
    big_array = store_arrays(root, [big_array, 'big_array'])[0]
  
  response = populate_array(self, 30000, 12, big_array, big_index, array_schema)
  message_list.extend(response['messages'])
  
  message_list.append('Write end: ' + get_process_memory_usage())
  return message_list

def bigger_than_memory_read(self, out_of_core_index=False):
  message_list = []
  message_list.append('Read start ' + get_process_memory_usage())
  
  root = self.getPhysicalRoot()
  
  # We don't have the columns names here, so we set manually the 14th column
  # name to 'quantity' to simplify things.
  columns = range(23)
  columns[13] = 'quantity'
  
  ob = root.big_index
  message_list.append(len(ob._p_jar.db().storage.load(ob._p_oid, '')[0])+42)
  
  big_index = root.big_index[:] if out_of_core_index else np.arange(360000)
  
  response = process_data(root.big_array[:], big_index[:], columns)
  message_list.extend(response['messages'])
  message_list.append(response['result'])

  message_list.append('Read end: ' + get_process_memory_usage())
  return message_list

def bigger_than_memory_profile(self):
  profiles = [bigger_than_memory_read_profile, bigger_than_memory_write_profile]
  return [profile(self) for profile in profiles]

def bigger_than_memory_read_profile(self):
  profile_path = '/tmp/profile_read'
  cProfile.runctx('bigger_than_memory_read(self)', globals(), locals(), profile_path)
  return "To see the profile start a console on server and: import pstats; pstats.Stats('%s').print_stats()" % profile_path
  
def bigger_than_memory_write_profile(self):
  profile_path = '/tmp/profile_write'
  cProfile.runctx('bigger_than_memory_write(self)', globals(), locals(), profile_path)
  return "To see the profile start a console on server and: import pstats; pstats.Stats('%s').print_stats()" % profile_path

@contextmanager
def timer(name, message_list):
  start_time = time()
  yield
  elapsed_time = time() - start_time
  message_list.append('[%s] finished in % seconds' % (name, elapsed_time))

def get_process_memory_usage():
  process = psutil.Process(os.getpid())
  m = process.memory_info()
  return 'VIRT: %i MB\tRSS: %iMB' % (m.vms//MB, m.rss//MB)

def get_field_names_and_schema():
  field_names = ['f%s' % number if not number == 13 else 'quantity' for number in range(23)]
  array_schema = np.dtype({
    'names' : field_names,
    'formats': map(
      np.dtype,
      ['i8', 'i8', 'i8', 'i8', 'i8', 'i8', 'i8', 'i8', 'i8', 'i8',
        'i8', 'i8', 'i8', 'f8', 'i8', 'i8', 'f8', 'f8', 'f8',
        'a50', 'a50', 'a50', 'a50']
        #'f8', 'f8', 'f8', 'f8']
    )
  })
  return field_names, array_schema

def create_zbigarray(dimension, dtype):
  return ZBigArray(dimension, dtype)
  
def store_arrays(root, *arrays_filenames):
  for array, filename in arrays_filenames:
    setattr(root, filename, None)
    setattr(root, filename, array)
  transaction.commit()
  return [getattr(root, filename) for _, filename in arrays_filenames]
  
def populate_array(root, chunk_size, number_of_chunks, big_array, big_index, schema):
  offset = 0
  start_populate_wendelin = mysql_time = time()
  # ZSQL Method to fetch data from stocks table, chunk by chunk, and put it in
  # the ZBigArray. Implemented using DTML's sqlvar instruction, LIMIT andd OFFSET.
  data = root.stock_offset(my_offset=0, my_limit=chunk_size)
  mysql_time = time() - mysql_time
  columns = data._names
  tuples = data.tuples()
  counter = 0
  message_list = []
  message_list.append('Running with %s chunks of %s rows.' % (number_of_chunks, chunk_size))
  
  chunk_byte_size = sys.getsizeof(tuples)
  message_list.append('Memory (in bytes) for each chunk: %s' % chunk_byte_size)
  for i in range(number_of_chunks):
    for index, row in enumerate(tuples):
      tuples[index] = tuple([filter_item(item) for item in row])
    if len(tuples) < chunk_size:
      chunk_end = chunk_size*(i) + len(tuples)
    else:
      chunk_end = chunk_size*(i+1)
    big_array[offset:chunk_end] = np.array(tuples, schema)
    offset += chunk_size
    temp_mysql_time = time()
    data = root.stock_offset(my_offset=offset, my_limit=chunk_size)
    mysql_time += time()-temp_mysql_time
    tuples = data.tuples()
  total_time = time() - start_populate_wendelin
  message_list.append('%s seconds to load & write %s records into wendelin array. %s seconds on MySQL loads and %s writting to disk.' % (total_time, chunk_end, mysql_time, total_time-mysql_time))
  transaction.commit()
  message_list.append('Total size of out-of-core array: %s megabytes.' % (big_array.nbytes // MB))
  return {'messages': message_list, 'columns': columns}

def filter_item(item):
  if not item or isinstance(item, type(None)): #or isinstance(item, (str, unicode)):
    return 0
  elif isinstance(item, DateTime):
    return 0
  else:
    return item

def process_data(big_array, big_index, columns):
  message_list = []
  result = None
  
  with timer('pandas sum', message_list):
    df = pd.DataFrame.from_records(big_array, index=big_index)
    result = df.quantity.sum()
  
  with timer('numpy sum', message_list):
    big_array['quantity'].sum()
  
  # message_list.append('%s seconds to read the wendelin array with %s rows.' % (finish-start, len(big_array)))
  # message_list.append('Total size of out-of-core array: %s megabytes.' % (big_array.nbytes // MB))
  return { 'messages': message_list, 'result': result }